# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import json
import os

import pytest


class AbstractFixtureLoader(object):
    def __init__(self, basedir):
        self.basedir = basedir

    def _read_file(self, test_dir, file_name):
        with open(os.path.join(self.basedir, test_dir, file_name)) as f:
            return f.read()

    def load_single(self, test_dir, file_name):
        return json.loads(self._read_file(test_dir, file_name))

    def load_test(self, test_dir):
        raise NotImplementedError('You have to implement me!')


@pytest.fixture()
def update_fixture_loader():
    class _Loader(AbstractFixtureLoader):

        def load_test(self, test_dir):
            root = self.load_single(test_dir, 'root.json')
            head = self.load_single(test_dir, 'head.json')
            update = self.load_single(test_dir, 'update.json')
            expected = self.load_single(test_dir, 'expected.json')
            desc = self._read_file(test_dir, 'description.txt')
            return root, head, update, expected, desc

    return _Loader('./tests/fixtures/update_scenarios/')



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

import copy
import logging

import six
from dictdiffer import ADD, CHANGE, REMOVE, patch
from dictdiffer.merge import Merger, UnresolvedConflictsException

from .config import DictMergerOps
from .conflict import Conflict, ConflictType
from .errors import MergeError
from .nothing import NOTHING
from .utils import (
    dedupe_list, del_obj_at_key_path, get_dotted_key_path, get_obj_at_key_path,
    set_obj_at_key_path
)

LOGGER = logging.getLogger(__name__)


def _get_list_fields(obj, res, key_path=()):
    if isinstance(obj, list):
        res.append(key_path)
    elif isinstance(obj, dict):
        for key, value in six.iteritems(obj):
            _get_list_fields(value, res, key_path + (key, ))

    return res


def patch_to_conflict_set(patch):
    """Translates a dictdiffer conflict into a json_merger one."""
    patch_type, patched_key, value = patch
    if isinstance(patched_key, list):
        key_path = tuple(patched_key)
    else:
        key_path = tuple(k for k in patched_key.split('.') if k)

    conflicts = set()
    if patch_type == REMOVE:
        conflict_type = ConflictType.REMOVE_FIELD
        for key, obj in value:
            conflicts.add(Conflict(conflict_type, key_path + (key, ), None))
    elif patch_type == CHANGE:
        conflict_type = ConflictType.SET_FIELD
        first_val, second_val = value
        conflicts.add(Conflict(conflict_type, key_path, second_val))
    elif patch_type == ADD:
        conflict_type = ConflictType.SET_FIELD
        for key, obj in value:
            conflicts.add(Conflict(conflict_type, key_path + (key, ), obj))

    return conflicts


class SkipListsMerger(object):
    """3-way Merger that ignores list fields."""

    def __init__(self, root, head, update, default_op,
                 data_lists=None, custom_ops={}, key_path=None):
        self.root = copy.deepcopy(root)
        self.head = copy.deepcopy(head)
        self.update = copy.deepcopy(update)
        self.custom_ops = custom_ops
        self.default_op = self._operation_to_function(default_op)
        self.data_lists = set(data_lists or [])
        self.key_path = key_path or []

        # We can have the same conflict appear more times because we keep only
        # one of the possible resolutions as a conflict while we apply the
        # other as a fallback. Sometimes multiple fallback dict diffs
        # conflict with a single change.
        self.conflict_set = set()
        self.skipped_lists = set()
        self.merged_root = None
        self.list_backups = {}

    def _build_skipped_lists(self):
        lists = set()
        lists.update(_get_list_fields(self.head, []))
        lists.intersection_update(_get_list_fields(self.update, []))
        for list_ in lists:
            dotted = get_dotted_key_path(list_, True)
            if dotted not in self.data_lists:
                self.skipped_lists.add(list_)

    def _backup_lists(self):
        self._build_skipped_lists()
        for list_ in self.skipped_lists:
            self.list_backups[list_] = (
                get_obj_at_key_path(self.root, list_),
                get_obj_at_key_path(self.head, list_),
                get_obj_at_key_path(self.update, list_))
            # The root is the only one that may not be there. Head and update
            # are retrieved using list intersection.
            del_obj_at_key_path(self.root, list_, False)
            del_obj_at_key_path(self.head, list_)
            del_obj_at_key_path(self.update, list_)

    def _restore_lists(self):
        for list_, (bak_r, bak_h, bak_u) in six.iteritems(self.list_backups):
            if bak_r is not None:
                set_obj_at_key_path(self.root, list_, bak_r)
            set_obj_at_key_path(self.head, list_, bak_h)
            set_obj_at_key_path(self.update, list_, bak_u)

    @property
    def conflicts(self):
        """List of conflicts for the current object."""

        # Make this compatible with the project convention (list of conflicts).
        return list(self.conflict_set)

    def _merge_base_values(self):
        if self.head == self.update:
            self.merged_root = self.head
        elif self.head == self.root:
            self.merged_root = self.update
        elif self.update == self.root:
            self.merged_root = self.head
        elif self.head == NOTHING:
            self.merged_root = self.update
        elif self.update == NOTHING:
            self.merged_root = self.head
        else:
            strategy = self._get_rule_for_field(self.key_path)
            self.merged_root, conflict = {
                'f': (self.head, self.update),
                's': (self.update, self.head)}[strategy]
            self.conflict_set.add(
                Conflict(ConflictType.SET_FIELD, (), conflict))

    def _merge_dicts(self):
        self._backup_lists()

        LOGGER.debug(
            "Merging dicts with root=%s, head=%s, update=%s",
            self.root,
            self.head,
            self.update,
        )
        non_list_merger = Merger(self.root, self.head, self.update, {})
        try:
            non_list_merger.run()
        except UnresolvedConflictsException as e:
            self._solve_dict_conflicts(non_list_merger, e.content)

        self._restore_lists()
        self.merged_root = patch(
            dedupe_list(non_list_merger.unified_patches),
            self.root
        )

    def _solve_dict_conflicts(self, non_list_merger, conflicts):
        strategies = [self._get_custom_strategy(conflict)
                      for conflict in conflicts]
        non_list_merger.continue_run(strategies)

        for conflict, strategy in zip(conflicts, strategies):
            conflict_patch = {'f': conflict.second_patch,
                              's': conflict.first_patch}[strategy]
            conflict_set = patch_to_conflict_set(conflict_patch)
            LOGGER.debug(
                "Solved conflict using strategy %s, conflicts=%s",
                strategy,
                conflict_set,
            )
            self.conflict_set.update(conflict_set)

    def _get_custom_strategy(self, patch):
        full_path = self._get_path_from_patch(patch)
        return self._get_rule_for_field(full_path)

    @staticmethod
    def _get_path_from_patch(patch):
        _, field, modification = patch.first_patch

        full_path = []
        if isinstance(field, list):
            full_path.extend(field)
        elif field:
            full_path.append(field)

        if (isinstance(modification, list) and modification[0][0]):
            full_path.append(modification[0][0])

        return full_path

    @staticmethod
    def _operation_to_function(operation):
        if callable(operation):
            return operation
        elif operation == DictMergerOps.FALLBACK_KEEP_HEAD:
            return lambda head, update, down_path: 'f'
        elif operation == DictMergerOps.FALLBACK_KEEP_UPDATE:
            return lambda head, update, down_path: 's'
        else:
            return lambda head, update, down_path: None

    def _absolute_path(self, field_path):
        full_path = list(self.key_path) + field_path
        return get_dotted_key_path(full_path, filter_int_keys=True)

    def _get_rule_for_field(self, field_path):
        current_path = self._absolute_path(field_path)
        head = get_obj_at_key_path(self.head, field_path)
        update = get_obj_at_key_path(self.update, field_path)
        down_path = []
        rule = None

        while current_path:
            operation = self._operation_to_function(
                self.custom_ops.get(current_path)
            )
            rule = operation(head, update, down_path)
            if rule:
                break
            current_path_parts = current_path.split('.')
            current_path = '.'.join(current_path_parts[:-1])
            down_path.append(current_path_parts[-1])

        return rule or self.default_op(head, update, field_path)

    def merge(self):
        """Perform merge of head and update starting from root."""
        if isinstance(self.head, dict) and isinstance(self.update, dict):
            if not isinstance(self.root, dict):
                self.root = {}
            self._merge_dicts()
        else:
            self._merge_base_values()

        if self.conflict_set:
            raise MergeError('Dictdiffer Errors', self.conflicts)



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


class DictMergerOps(object):
    """Possible strategies for merging two base values.

    Attributes:
        FALLBACK_KEEP_HEAD: In case of conflict keep the `head` value.

        FALLBACK_KEEP_UPDATE: In case of conflict keep the `update` value.
    """
    allowed_ops = [
        'FALLBACK_KEEP_HEAD',
        'FALLBACK_KEEP_UPDATE'
    ]

    @staticmethod
    def keep_longest(head, update, down_path):
        """Keep longest field among `head` and `update`.
        """
        if update is None:
            return 'f'
        if head is None:
            return 's'

        return 'f' if len(head) >= len(update) else 's'


for mode in DictMergerOps.allowed_ops:
    setattr(DictMergerOps, mode, mode)


class UnifierOps(object):
    """
    Attributes:
        KEEP_ONLY_HEAD_ENTITIES: Merge entities in `update` with their match
            in `head` having as a base the match in `root`.

        KEEP_ONLY_UPDATE_ENTITIES: Merge entities in 'head' with their match
            in `update` having as a base the match in `root`.

        KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST: Perform an union of all
            entities from `head` and `update` and merge the matching ones.
            Also, preserve the order relations between the entities in both
            lists. If two entities can have the same position first pick the
            one that is present in the `head` object.

        KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST: Same behavior as
            KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST but first pick the
            `update` entities.

        KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE: If an entity was added
            in the diff between the `root` and `head` lists but it's not
            present in the `update` list then raise a conflict.
    """
    allowed_ops = [
        'KEEP_ONLY_HEAD_ENTITIES',
        'KEEP_ONLY_UPDATE_ENTITIES',
        'KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST',
        'KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST',
        'KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE',
        'KEEP_UPDATE_AND_HEAD_ENTITIES_CONFLICT_ON_HEAD_DELETE',
        'KEEP_HEAD_ENTITIES_CONFLICT_ON_NEW_UPDATE'
    ]


for mode in UnifierOps.allowed_ops:
    setattr(UnifierOps, mode, mode)



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Definition for JSON merger class."""

from __future__ import absolute_import, print_function

import copy
import logging

from .comparator import DefaultComparator
from .dict_merger import SkipListsMerger
from .errors import MergeError
from .list_unify import ListUnifier
from .nothing import NOTHING
from .utils import (
    get_conf_set_for_key_path, get_dotted_key_path, get_obj_at_key_path,
    set_obj_at_key_path
)

PLACEHOLDER_STR = '#$PLACEHOLDER$#'
LOGGER = logging.getLogger(__name__)


class Merger(object):
    """Class that merges two JSON objects that share a common ancestor.

    This class treats by default all lists as being lists of entities and
    offers support for matching their elements by their content, by specifing
    per-field comparator classes.
    """

    def __init__(self, root, head, update,
                 default_dict_merge_op, default_list_merge_op,
                 list_dict_ops=None, list_merge_ops=None,
                 comparators=None, data_lists=None):
        """
        Args:
            root: A common ancestor of the two objects being merged.

            head: One of the objects that is being merged. Refers to the
                version that is currently in use. (e.g. a displayed database
                record)

            update: The second object that is being merged. Refers to an update
                that needs to be integrated with the in-use version.

            default_dict_merge_op
              (:class:`json_merger.config.DictMergerOps` class attribute):
                Default strategy for merging regular non list JSON values
                (strings, numbers, other objects).

            default_list_merge_op
              (:class:`json_merger.config.UnifierOps` class attribute):
                Default strategy for merging two lists of entities.

            dict_merge_ops: Defines custom strategies for merging dict of
                entities.

                Dict formatted as:
                    * keys -- a config string
                    * values -- a class attribute of
                      :class:`json_merger.config.DictMergerOps`

            list_merge_ops: Defines custom strategies for merging lists of
                entities.

                Dict formatted as:
                    * keys -- a config string
                    * values -- a class attribute of
                      :class:`json_merger.config.UnifierOps`

            comparators: Defines classes used for rendering entities in list
                fields as equal.

                Dict formatted as:
                    * keys -- a config string
                    * values -- a class that extends
                      :class:`json_merger.comparator.BaseComparator`

            data_lists: List of config strings defining the lists that are not
                treated as lists of entities.

        Note:
            A configuration string represents the path towards a list field in
            the object sepparated with dots.

        Example:
            Configuration strings can be:

                For ``{'lst': [{'id': 0, 'tags': ['foo', 'bar']}]}``:

                * the config string for the top level list is ``'lst'``
                * the config string for the tags lists is ``'lst.tags'``
        """
        self.comparators = comparators or {}
        self.data_lists = set(data_lists or [])
        self.list_dict_ops = list_dict_ops or {}
        self.list_merge_ops = list_merge_ops or {}

        self.default_dict_merge_op = default_dict_merge_op
        self.default_list_merge_op = default_list_merge_op

        self.root = copy.deepcopy(root)
        self.head = copy.deepcopy(head)
        self.update = copy.deepcopy(update)
        self.head_stats = {}
        self.update_stats = {}

        self.conflicts = []
        self.merged_root = None

        self.aligned_root = copy.deepcopy(root)
        self.aligned_head = copy.deepcopy(head)
        self.aligned_update = copy.deepcopy(update)

    def merge(self):
        """Populates result members.

        Performs the merge algorithm using the specified config and fills in
        the members that provide stats about the merging procedure.

        Attributes:
            merged_root: The result of the merge.

            aligned_root, aligned_head, aligned_update: Copies of root, head
                and update in which all matched list entities have the same
                list index for easier diff viewing.

            head_stats, update_stats: Stats for each list field present in the
                head or update objects. Instance of
                :class:`json_merger.stats.ListMatchStats`

            conflicts: List of :class:`json_merger.conflict.Conflict` instances
                that occured during the merge.

        Raises:
            :class:`json_merger.errors.MergeError` : If conflicts occur during
                the call.

        Example:
            >>> from json_merger import Merger
            >>> # We compare people by their name
            >>> from json_merger.comparator import PrimaryKeyComparator
            >>> from json_merger.config import DictMergerOps, UnifierOps
            >>> from json_merger.errors import MergeError
            >>> # Use this only for doctest :)
            >>> from pprint import pprint as pp
            >>>
            >>> root = {'people': [{'name': 'Jimmy', 'age': 30}]}
            >>> head = {'people': [{'name': 'Jimmy', 'age': 31},
            ...                    {'name': 'George'}]}
            >>> update = {'people': [{'name': 'John'},
            ...                      {'name': 'Jimmy', 'age': 32}]}
            >>>
            >>> class NameComparator(PrimaryKeyComparator):
            ...     # Two objects are the same entitity if they have the
            ...     # same name.
            ...     primary_key_fields = ['name']
            >>> m = Merger(root, head, update,
            ...            DictMergerOps.FALLBACK_KEEP_HEAD,
            ...            UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST,
            ...            comparators = {'people': NameComparator})
            >>> # We do a merge
            >>> try:
            ...     m.merge()
            ... except MergeError as e:
            ...     # Conflicts are the same thing as the exception content.
            ...     assert e.content == m.conflicts
            >>> # This is how the lists are aligned:
            >>> pp(m.aligned_root['people'], width=60)
            ['#$PLACEHOLDER$#',
             {'age': 30, 'name': 'Jimmy'},
             '#$PLACEHOLDER$#']
            >>> pp(m.aligned_head['people'], width=60)
            ['#$PLACEHOLDER$#',
             {'age': 31, 'name': 'Jimmy'},
             {'name': 'George'}]
            >>> pp(m.aligned_update['people'], width=60)
            [{'name': 'John'},
             {'age': 32, 'name': 'Jimmy'},
             '#$PLACEHOLDER$#']
            >>> # This is the end result of the merge:
            >>> pp(m.merged_root, width=60)
            {'people': [{'name': 'John'},
                        {'age': 31, 'name': 'Jimmy'},
                        {'name': 'George'}]}
            >>> # With some conflicts:
            >>> pp(m.conflicts, width=60)
            [('SET_FIELD', ('people', 1, 'age'), 32)]
            >>> # And some stats:
            >>> pp(m.head_stats[('people',)].in_result)
            [{'age': 31, 'name': 'Jimmy'}, {'name': 'George'}]
            >>> pp(m.update_stats[('people',)].not_in_result)
            []

        Note:
            Even if conflicts occur, merged_root, aligned_root, aligned_head
            and aligned_update are always populated by following the
            startegies set for the merger instance.
        """
        self.merged_root = self._recursive_merge(self.root, self.head,
                                                 self.update)
        if self.conflicts:
            raise MergeError('Conflicts Occurred in Merge Process',
                             self.conflicts)

    def _recursive_merge(self, root, head, update, key_path=()):
        dotted_key_path = get_dotted_key_path(key_path, filter_int_keys=True)

        if (isinstance(head, list) and isinstance(update, list) and
                dotted_key_path not in self.data_lists):
            # In this case we are merging two lists of objects.
            lists_to_unify = [()]
            if not isinstance(root, list):
                root = []
        else:
            # Otherwise we merge everything but the lists using DictMergerOps.
            m = self._merge_objects(root, head, update, key_path)
            root = m.merged_root
            lists_to_unify = m.skipped_lists

        for list_field in lists_to_unify:
            absolute_key_path = key_path + list_field

            root_l = get_obj_at_key_path(root, list_field, [])
            head_l = get_obj_at_key_path(head, list_field, [])
            update_l = get_obj_at_key_path(update, list_field, [])

            unifier = self._unify_lists(root_l, head_l, update_l,
                                        absolute_key_path)

            new_list = []
            for idx, objects in enumerate(unifier.unified):
                root_obj, head_obj, update_obj = objects
                LOGGER.debug(
                    "Merging matched elements: root=%s, head=%s, update=%s",
                    root_obj,
                    head_obj,
                    update_obj
                )
                new_obj = self._recursive_merge(root_obj, head_obj, update_obj,
                                                absolute_key_path + (idx, ))
                if new_obj != NOTHING:
                    new_list.append(new_obj)

            root = set_obj_at_key_path(root, list_field, new_list)
            self._build_aligned_lists_and_stats(unifier, absolute_key_path)

        return root

    def _merge_objects(self, root, head, update, key_path):
        data_lists = get_conf_set_for_key_path(self.data_lists, key_path)

        LOGGER.debug("Merging non-lists at %s", key_path)

        object_merger = SkipListsMerger(root, head, update,
                                        self.default_dict_merge_op,
                                        data_lists, self.list_dict_ops,
                                        key_path)

        try:
            object_merger.merge()
        except MergeError as e:
            self.conflicts.extend(c.with_prefix(key_path) for c in e.content)

        return object_merger

    def _unify_lists(self, root, head, update, key_path):
        dotted_key_path = get_dotted_key_path(key_path, True)

        operation = self.list_merge_ops.get(dotted_key_path,
                                            self.default_list_merge_op)
        comparator_cls = self.comparators.get(dotted_key_path,
                                              DefaultComparator)

        LOGGER.debug(
            "Unifying lists at %s using operation %s and comparator %s",
            key_path,
            operation,
            comparator_cls,
        )
        list_unifier = ListUnifier(root, head, update,
                                   operation, comparator_cls)
        try:
            list_unifier.unify()
        except MergeError as e:
            self.conflicts.extend(c.with_prefix(key_path) for c in e.content)

        return list_unifier

    def _build_aligned_lists_and_stats(self, list_unifier, key_path):
        root_list = []
        head_list = []
        update_list = []
        for root_obj, head_obj, update_obj in list_unifier.unified:
            # Cast NOTHING objects to a placeholder so we reserialize back to
            # JSON if needed.
            root_list.append(root_obj or PLACEHOLDER_STR)
            head_list.append(head_obj or PLACEHOLDER_STR)
            update_list.append(update_obj or PLACEHOLDER_STR)

        # Try to put back the list if the key path existed in the first place.
        self.aligned_root = set_obj_at_key_path(self.aligned_root,
                                                key_path, root_list, False)
        self.aligned_head = set_obj_at_key_path(self.aligned_head,
                                                key_path, head_list, False)
        self.aligned_update = set_obj_at_key_path(self.aligned_update,
                                                  key_path, update_list, False)

        # Also copy over the stats.
        self.head_stats[key_path] = list_unifier.head_stats
        self.update_stats[key_path] = list_unifier.update_stats



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Test merger corner cases that are unlikely to appear as usage scenarios."""

from __future__ import absolute_import, print_function


import pytest


from json_merger.config import DictMergerOps, UnifierOps
from json_merger.conflict import Conflict, ConflictType
from json_merger.errors import MergeError
from json_merger.merger import Merger
from json_merger.dict_merger import patch_to_conflict_set


def test_merge_bare_int_lists():
    r = [1, 2, 3]
    h = [1, 2, 3, 4]
    u = [1, 2, 5]

    m = Merger(r, h, u,
               DictMergerOps.FALLBACK_KEEP_HEAD,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    m.merge()
    assert m.merged_root == [1, 2, 5]


def test_merge_bare_str_lists():
    r = ['1', '2', '3']
    h = ['1', '2', '3', '4']
    u = ['1', '2', '5']

    m = Merger(r, h, u,
               DictMergerOps.FALLBACK_KEEP_HEAD,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    m.merge()
    assert m.merged_root == ['1', '2', '5']


def test_merge_nested_lists():
    r = [[1], [2], [3]]
    h = [[1], [2], [3], [4]]
    u = [[1], [2], [5]]

    m = Merger(r, h, u,
               DictMergerOps.FALLBACK_KEEP_HEAD,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    m.merge()

    assert m.merged_root == [[1], [2], [5]]


def test_merge_root_is_not_list():
    r = 'randomstring'
    h = [[1], [2, 3], [5]]
    u = [[1], [2, 3], [5]]

    m = Merger(r, h, u,
               DictMergerOps.FALLBACK_KEEP_HEAD,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    m.merge()
    # Here the lists are aligned as entities and lists of entities.
    assert m.merged_root == [[1], [2, 3], [5]]


def test_merge_list_with_string():
    r = 'somerandomvalue'
    h = [1, 2, 3]
    u = 'a given string'

    m = Merger(r, h, u,
               DictMergerOps.FALLBACK_KEEP_HEAD,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    with pytest.raises(MergeError) as excinfo:
        m.merge()

    assert m.merged_root == [1, 2, 3]
    assert len(excinfo.value.content) == 1
    assert excinfo.value.content[0] == Conflict(ConflictType.SET_FIELD, (),
                                                'a given string')


def test_merge_str_with_keep_longest():
    r = {}
    h = 'A short string'
    u = 'A much longer string'

    m = Merger(r, h, u,
               DictMergerOps.keep_longest,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)

    with pytest.raises(MergeError):
        m.merge()

    assert m.merged_root == 'A much longer string'


def test_merge_dict_with_keep_longest():
    r = {}
    h = {
        'a': 'A short string',
        'b': 'An extremely long string',
    }
    u = {
        'a': 'A much longer string',
        'b': 'Another short string',
        'c': 'Other string',
    }

    m = Merger(r, h, u,
               DictMergerOps.keep_longest,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)

    expected = {
        'a': 'A much longer string',
        'b': 'An extremely long string',
        'c': 'Other string',
    }

    with pytest.raises(MergeError):
        m.merge()

    assert m.merged_root == expected


def test_merge_list_with_keep_longest():
    r = {
        'a': [
            {
                'b': 'One string',
            },
        ],
    }
    h = {
        'a': [
            {
                'b': 'A different string',
            },
        ],
    }
    u = {}

    m = Merger(r, h, u,
               DictMergerOps.keep_longest,
               UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)

    expected = {
        'a': [
            {
                'b': 'A different string',
            },
        ],
    }  # as len(h['a']) > len(u['a']) and strategy is keep_longest

    with pytest.raises(MergeError):
        m.merge()

    assert m.merged_root == expected


def test_patch_to_conflict_set_handles_change_patch_with_dotted_key():
    patch = ('change', 'a.b', ('1', '2'))

    conflicts = patch_to_conflict_set(patch)

    expected_conflicts = {
        ('SET_FIELD', ('a', 'b'), '2'),
    }

    assert conflicts == expected_conflicts


def test_patch_to_conflict_set_handles_change_patch_with_list_key():
    patch = ('change', ['a', 0, 'b'], ('1', '2'))

    conflicts = patch_to_conflict_set(patch)

    expected_conflicts = {
        ('SET_FIELD', ('a', 0, 'b'), '2'),
    }

    assert conflicts == expected_conflicts


def test_keep_head_conflict_on_new_update():
    r = [1]
    h = [1, 2]
    u = [3]

    m = Merger(r, h, u,
               DictMergerOps.keep_longest,
               UnifierOps.KEEP_HEAD_ENTITIES_CONFLICT_ON_NEW_UPDATE)
    with pytest.raises(MergeError):
        m.merge()

    expected_merge = [2]
    expected_conflict = [('INSERT', (0,), 3)]
    assert m.merged_root == expected_merge
    assert m.conflicts == expected_conflict


def test_merger_doesnt_add_elements_removed_in_head_when_root_matches_update():
    update = {
        'ids': [
            {'schema': 'INSPIRE ID', 'value': 'INSPIRE-00231894'},
            {'schema': 'ORCID', 'value': '0000-0002-0165-6297'}
        ]
    }
    head = {
        'ids': [
            {'value': 'INSPIRE-00231894', 'schema': 'INSPIRE ID'},
            {'value': 'V.Nikolaenko.1', 'schema': 'INSPIRE BAI'}
        ]
    }
    root = {
        'ids': [
            {'schema': 'INSPIRE ID', 'value': 'INSPIRE-00231894'},
            {'schema': 'ORCID', 'value': '0000-0002-0165-6297'}
        ]
    }

    m = Merger(root, head, update,
               DictMergerOps.FALLBACK_KEEP_HEAD,
               UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST)

    m.merge()
    assert m.merged_root['ids'] == [
        {'value': 'INSPIRE-00231894', 'schema': 'INSPIRE ID'},
        {'value': 'V.Nikolaenko.1', 'schema': 'INSPIRE BAI'}
    ]



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Python module that is able to merge json record objects."""

import sys

from setuptools import find_packages, setup

readme = open("README.rst").read()
history = open("CHANGES.rst").read()


url = "https://github.com/inveniosoftware-contrib/json-merger"


tests_require = [
    "check-manifest>=0.25",
    "coverage>=4.0",
    "isort>=4.2.2",
    "pydocstyle>=1.0.0",
    "pytest-cache>=1.0",
    "pytest-cov>=1.8.0",
    "pytest>=4.6.11",
    "flake8>=3.9.0",
]

contrib_require = [
    "editdistance>=0.3.1",
    "munkres<=1.0.12",
    "Unidecode==0.4.19" if sys.version_info < (3, 6) else "Unidecode>=0.4.19",
]

tests_require += contrib_require

extras_require = {
    "contrib": contrib_require,
    "docs": [
        "Sphinx>=1.4.2",
    ],
    "tests": tests_require,
}

extras_require["all"] = []
for reqs in extras_require.values():
    extras_require["all"].extend(reqs)

install_requires = [
    "autosemver==0.5.5",
    "dictdiffer==0.8.1" if sys.version_info < (3, 6) else "dictdiffer>=0.6.0",
    "six>=1.10.0",
    "pyrsistent>=0.11.13",
]

setup_requires = [
    "autosemver==0.5.5",
]

packages = find_packages()

setup(
    name="json-merger",
    autosemver=True,
    description=__doc__,
    long_description=readme + "\n\n" + history,
    keywords="JSON patch merge conflict",
    license="GPLv2",
    author="CERN",
    author_email="admin@inspirehep.net",
    url=url,
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    entry_points={},
    setup_requires=setup_requires,
    extras_require=extras_require,
    install_requires=install_requires,
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Development Status :: 4 - Beta",
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Module for aligning the same entities in lists."""

from __future__ import absolute_import, print_function

from .comparator import DefaultComparator
from .config import UnifierOps
from .conflict import Conflict, ConflictType
from .errors import MergeError
from .graph_builder import (
    ListMatchGraphBuilder, sort_cyclic_graph_best_effort, toposort
)
from .nothing import Nothing

_SOURCES = {
    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES: ['update'],
    UnifierOps.KEEP_ONLY_HEAD_ENTITIES: ['head'],
    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST: ['update', 'head'],
    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST: ['update', 'head'],
    UnifierOps.KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE: ['update'],
    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_CONFLICT_ON_HEAD_DELETE:
        ['update', 'head'],
    UnifierOps.KEEP_HEAD_ENTITIES_CONFLICT_ON_NEW_UPDATE: ['head', 'update'],
}

_PICK_FIRST = {
    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES: 'update',
    UnifierOps.KEEP_ONLY_HEAD_ENTITIES: 'head',
    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST: 'head',
    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST: 'update',
    UnifierOps.KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE: 'update',
    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_CONFLICT_ON_HEAD_DELETE: 'update',
    UnifierOps.KEEP_HEAD_ENTITIES_CONFLICT_ON_NEW_UPDATE: 'update',
}

_RAISE_ERROR_OPS = [
    UnifierOps.KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE,
    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_CONFLICT_ON_HEAD_DELETE,
]

_RAISE_ON_UPDATE_CHANGED = [
    UnifierOps.KEEP_HEAD_ENTITIES_CONFLICT_ON_NEW_UPDATE
]


class ListUnifier(object):

    def __init__(self, root, head, update, operation,
                 comparator_cls=DefaultComparator):
        if operation not in UnifierOps.allowed_ops:
            raise ValueError('Operation %r not permitted' % operation)

        self.root = root
        self.head = head
        self.update = update
        self.comparator_cls = comparator_cls

        self.head_stats = None
        self.update_stats = None

        # Whether to raise error on deleting a head entity.
        self.raise_on_head_delete = operation in _RAISE_ERROR_OPS
        # Whether to raise on new entity in update
        self.raise_on_new_update = operation in _RAISE_ON_UPDATE_CHANGED
        # Sources from which to keep entities.
        self.sources = _SOURCES[operation]
        # Source from which to pick the first element when they can be
        # interchanged in the topological sort.
        self.pick_first = _PICK_FIRST[operation]

        self.unified = []

    def unify(self):
        graph_builder = ListMatchGraphBuilder(
            self.root, self.head, self.update, self.sources,
            self.comparator_cls)
        graph, nodes = graph_builder.build_graph()
        self.head_stats = graph_builder.head_stats
        self.update_stats = graph_builder.update_stats

        conflicts = []
        if graph_builder.multiple_match_choices:
            conflicts = [Conflict(ConflictType.MANUAL_MERGE, (), choice)
                         for choice in graph_builder.multiple_match_choices]

        try:
            node_order = toposort(graph, self.pick_first)
        except ValueError:
            node_order = sort_cyclic_graph_best_effort(graph, self.pick_first)
            conflicts.append(Conflict(ConflictType.REORDER, (), None))

        for node in node_order:
            self.unified.append(nodes[node])
        if (self.raise_on_head_delete and
                self.head_stats.not_in_result):
            removed = self.head_stats.not_in_result
            conflicts.extend([Conflict(ConflictType.ADD_BACK_TO_HEAD, (), r)
                              for r in removed])
        if self.raise_on_new_update:
            idx_to_remove = []
            for idx, (root, head, update) in enumerate(self.unified):
                if isinstance(root, Nothing) and \
                        isinstance(head, Nothing) and \
                        not isinstance(update, Nothing):
                    conflicts.append(
                        Conflict(ConflictType.INSERT, (idx,), update)
                    )
                    idx_to_remove.append(idx)
            for idx in sorted(idx_to_remove, reverse=True):
                del self.unified[idx]
        if conflicts:
            raise MergeError('Errors in list unifier', conflicts)



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

import re

import editdistance
import six

from munkres import Munkres
from unidecode import unidecode

_RE_NAME_TOKEN_SEPARATOR = re.compile(r'[^\w\'-]+', re.UNICODE)


def _normalized_edit_dist(s1, s2):
    return float(editdistance.eval(s1, s2)) / max(len(s1), len(s2), 1)


class NameToken(object):
    def __init__(self, token):
        self.token = token.lower()

    def __eq__(self, other):
        return self.token == other.token

    def __repr__(self):
        return repr(u'{}: {}'.format(self.__class__.__name__,
                                     self.token))


class NameInitial(NameToken):
    def __eq__(self, other):
        return self.token == other.token[:len(self.token)]


def token_distance(t1, t2, initial_match_penalization):
    """Calculates the edit distance between two tokens."""
    if isinstance(t1, NameInitial) or isinstance(t2, NameInitial):
        if t1.token == t2.token:
            return 0
        if t1 == t2:
            return initial_match_penalization
        return 1.0
    return _normalized_edit_dist(t1.token, t2.token)


def simple_tokenize(name):
    """Simple tokenizer function to be used with the normalizers."""
    last_names, first_names = name.split(',')
    last_names = _RE_NAME_TOKEN_SEPARATOR.split(last_names)
    first_names = _RE_NAME_TOKEN_SEPARATOR.split(first_names)

    first_names = [NameToken(n) if len(n) > 1 else NameInitial(n)
                   for n in first_names if n]
    last_names = [NameToken(n) if len(n) > 1 else NameInitial(n)
                  for n in last_names if n]
    return {'lastnames': last_names,
            'nonlastnames': first_names}


class AuthorNameNormalizer(object):
    """Callable that normalizes an author name given a tokenizer function."""

    def __init__(self, tokenize_function,
                 first_names_number=None,
                 first_name_to_initial=False,
                 asciify=False):
        """Initialize the normalizer.

        Args:
            tokenize_function:
                A function that receives an author name and parses it out in
                the following format:
                    {'lastnames': NameToken instance list,
                     'nonlastnames': NameToken instance list}
            first_names_number:
                Max number of first names too keep in the normalized name.
                If None, keep all first names
            first_name_to_initial:
                If set to True, all first names will be transformed into
                initials.
            asciify:
                If set to True, all non-ASCII characters will be replaced by
                the closest ASCII character, e.g. 'é' -> 'e'.
        """

        self.tokenize_function = tokenize_function
        self.first_names_number = first_names_number
        self.first_name_to_initial = first_name_to_initial
        self.normalize_chars = lambda x: _asciify(x) if asciify else x

    def __call__(self, author):
        name = author.get('full_name', '')
        name = _decode_if_not_unicode(name)
        name = self.normalize_chars(name)
        tokens = self.tokenize_function(name)
        last_fn_char = 1 if self.first_name_to_initial else None
        last_fn_idx = self.first_names_number

        return (tuple(n.token.lower() for n in tokens['lastnames']) +
                tuple(n.token.lower()[:last_fn_char]
                      for n in tokens['nonlastnames'][:last_fn_idx]))


class AuthorNameDistanceCalculator(object):
    """Callable that calculates a distance between two author's names."""

    def __init__(self, tokenize_function, match_on_initial_penalization=0.05,
                 full_name_field='full_name'):
        """Initialize the distance calculator.

        Args:
            tokenize_function: A function that receives an author name and
                parses it out in the following format:
                  {'lastnames': NameToken instance list,
                   'nonlastnames': NameToken instance list}
            match_on_initial_penalization:
                The cost value of a match between an initial and a full name
                starting with the same letter.
            name_field:
                The field in which an author record keeps the full name.
        Note:
            The default match_on_initial_penalization had the best results
            on a test suite based on production data.
        """
        self.tokenize_function = tokenize_function
        self.match_on_initial_penalization = match_on_initial_penalization
        self.name_field = full_name_field

    def __call__(self, author1, author2):
        # Return 1.0 on missing features.
        if self.name_field not in author1:
            return 1.0
        if self.name_field not in author2:
            return 1.0

        # Normalize to unicode
        name_a1 = _asciify(_decode_if_not_unicode(author1[self.name_field]))
        name_a2 = _asciify(_decode_if_not_unicode(author2[self.name_field]))

        tokens_a1 = self.tokenize_function(name_a1)
        tokens_a2 = self.tokenize_function(name_a2)
        tokens_a1 = tokens_a1['lastnames'] + tokens_a1['nonlastnames']
        tokens_a2 = tokens_a2['lastnames'] + tokens_a2['nonlastnames']

        # Match all names by editdistance.
        dist_matrix = [
            [token_distance(t1, t2, self.match_on_initial_penalization)
             for t2 in tokens_a2] for t1 in tokens_a1]

        matcher = Munkres()
        indices = matcher.compute(dist_matrix)
        cost = 0.0
        matched_only_initials = True
        for idx_a1, idx_a2 in indices:
            cost += dist_matrix[idx_a1][idx_a2]
            if (not isinstance(tokens_a1[idx_a1], NameInitial) or
                    not isinstance(tokens_a2[idx_a2], NameInitial)):
                matched_only_initials = False

        # Johnny, D will not be equal with Donny, J
        if matched_only_initials:
            return 1.0

        return cost / max(min(len(tokens_a1), len(tokens_a2)), 1.0)


def _decode_if_not_unicode(value):
    to_return = value

    if not isinstance(value, six.text_type):
        to_return = value.decode('utf-8')

    return to_return


def _asciify(value):
    return six.text_type(unidecode(value))
